"""
Welcome to gmx-top4py!
"""

from __future__ import annotations
import warnings

warnings.simplefilter("once", DeprecationWarning)
warnings.warn(
    "gmx-top4py has been renamed to gmxtop. Use 'pip install gmxtop'.",
    DeprecationWarning,
)
